import numpy as np
import scipy as sp


class TiseSolverFgh1D:
    def __init__(
        self,
        xa: float,
        xb: float,
        nx: int,
        num_eig: int,
    ):
        self.xa: float = xa
        self.xb: float = xb
        self.nx: int = nx
        x = np.linspace(xa, xb, nx, endpoint=False, dtype=np.float64)
        self.x: np.ndarray = x

        self.Vx: np.ndarray = self.V(x)
        self.num_eig: int = num_eig

        self.eigenvalues = np.zeros(self.num_eig, dtype=np.complex128)
        self.eigenvectors = np.zeros((self.num_eig, self.nx), dtype=np.complex128)

    # TISE potential
    def V(self, x: np.array):
        V = 0.5 * x * x
        return V

    def abscissa(self):
        return self.x

    def solve(self):
        # Hamiltonian matrix (N x N)
        matH = TiseSolverFgh1D.matH(self.x, self.Vx)

        # Diagonalize
        is_hermite: bool = False
        is_full_diagonal: bool = self.num_eig == self.nx
        if is_full_diagonal and is_hermite:
            eigenvalues, eigenvectors = sp.linalg.eig(matH, left=False, right=True)
        elif is_full_diagonal and (not is_hermite):
            eigenvalues, eigenvectors = sp.linalg.eigh(matH, left=False, right=True)
        elif (not is_full_diagonal) and is_hermite:
            eigenvalues, eigenvectors = sp.sparse.linalg.eigsh(matH, k=self.num_eig, which="SA")
        else:  # (not is_full_diagonal) and is_hermite:
            eigenvalues, eigenvectors = sp.sparse.linalg.eigs(matH, k=self.num_eig, which="SR")

        # Sort eigenvalues and eigenvectors in assending order by real part of the eigenvalues.
        idx = np.argsort(eigenvalues.real)
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[:, idx]

        # Normalize to \int |\psi(x)|^2 dx = 1, (Just in case)
        S = np.sum(np.abs(eigenvectors) ** 2, axis=0)
        eigenvectors /= np.sqrt(S)

        # Change eigenvectors
        eigenvectors = eigenvectors.T

        self.eigenvalues[:] = eigenvalues[:].copy()
        self.eigenvectors[:, :] = eigenvectors[:, :].copy()

        return eigenvalues

    @staticmethod
    def T2N_legacy(n, N):
        if N % 2 == 0:
            if n == 0:
                T = (1.0 / 12.0) * (N * N + 2.0)
            else:
                T = 0.5 * (-1.0) ** n / ((np.sin(np.pi * n / N)) ** 2)
        else:
            if n == 0:
                T = (1.0 / 12.0) * (N**2 - 1.0)
            else:
                T = 0.5 * (-1.0) ** n * np.cos(np.pi * n / N) / ((np.sin(np.pi * n / N)) ** 2)

        return T

    @staticmethod
    def T2N(n, N):
        """
        ベクトル n (整数値の配列) に対して T2N(n, N) を要素ごとに計算する関数
        """
        n = np.asarray(n)
        result = np.empty_like(n, dtype=np.float64)
        mask = n == 0
        if N % 2 == 0:
            T0 = (1.0 / 12.0) * (N * N + 2.0)
            result[mask] = T0
            nonzero = ~mask
            result[nonzero] = 0.5 * ((-1) ** np.abs(n[nonzero])) / (np.sin(np.pi * n[nonzero] / N) ** 2)
        else:
            T0 = (1.0 / 12.0) * (N**2 - 1)
            result[mask] = T0
            nonzero = ~mask
            result[nonzero] = 0.5 * ((-1) ** np.abs(n[nonzero])) * np.cos(np.pi * n[nonzero] / N) / (np.sin(np.pi * n[nonzero] / N) ** 2)
        return result

    @staticmethod
    def matH(
        x: np.ndarray,
        V: np.ndarray,
        is_code_optimization: bool = True,
    ):
        Nx = x.shape[0]  # = V.shape[0]
        dx = x[1] - x[0]
        Kx = 2.0 * np.pi / dx

        # Initialize the Hamiltonian matrix
        ham = np.zeros((Nx, Nx), dtype=np.float64)

        # Calculate grid spacing and wave vector

        if is_code_optimization:
            # x方向の運動エネルギー項: ブロック対角行列 np.kron(np.eye(Ny), T_x)
            n_idx = np.arange(Nx)
            diff_x = n_idx[:, None] - n_idx[None, :]  # shape: (Nx, Nx)
            T_x = TiseSolverFgh1D.T2N(diff_x, Nx)
            ham[:, :] = 0.5 * (Kx / Nx) ** 2 * T_x

            # 対角要素を追加
            ham[np.diag_indices(Nx)] += V
        else:
            # Legacy code (not optimized, 20~50 times slower than optimized code)
            if Nx % 2 == 0:
                for i in range(Nx):
                    for j in range(Nx):
                        if j == i:
                            ham[i, j] = 0.5 * ((Nx**2 + 2) / 12.0) * (Kx / Nx) ** 2 + V[i]
                        else:
                            ham[i, j] = 0.5 * (0.5 * (-1) ** (i - j) / ((np.sin(np.pi * (i - j) / Nx)) ** 2)) * (Kx / Nx) ** 2
            else:
                for i in range(Nx):
                    for j in range(Nx):
                        if j == i:
                            ham[i, j] = 0.5 * ((Nx**2 - 1) / 12.0) * (Kx / Nx) ** 2 + V[i]
                        else:
                            ham[i, j] = (
                                0.5 * (0.5 * (-1) ** (i - j) * np.cos(np.pi * (i - j) / Nx) / ((np.sin(np.pi * (i - j) / Nx)) ** 2)) * (Kx / Nx) ** 2
                            )

        return ham

    def eigenfunctions_scaling(self, x_new: np.ndarray):

        if self.x.shape == x_new.shape:
            rtol = atol = 10.0 * np.finfo(np.float64).eps
            if np.allclose(self.x, x_new, rtol=rtol, atol=atol):
                # abscissa is same as fgh-grid
                eigenfunctions = self.eigenvectors * np.sqrt(self.nx / (self.xb - self.xa))
                return eigenfunctions

        # Interpolation using FGH basis
        Lx = self.xb - self.xa
        # --- x-direction basis functions ---
        # \varphi(n,x) = \frac{1}{\sqrt{NL}}\sum_l e^{i k_l (x-x_n)}
        #              = \frac{1}{\sqrt{NL}}\sum_l e^{i 2\pi l (x/L-n/N)}
        if self.nx % 2 == 0:
            l_idx_x = np.arange(-self.nx // 2, self.nx // 2)  # shape: (nx,)
        else:
            l_idx_x = np.arange(-(self.nx - 1) // 2, (self.nx - 1) // 2 + 1)
        n_vals = np.arange(self.nx)  # basis index n, shape: (nx,)
        n_over = n_vals / self.nx  # shape: (nx,)
        # Normalized x-coordinates (shape: (len(x_new),))
        A = (x_new - self.xa) / Lx
        # For each n, compute A - n/nx, shape: (nx, len(x_new))
        temp_x = A[None, :] - n_over[:, None]
        # Compute exponent for each n, x, and l, resulting shape: (nx, len(x_new), nx)
        # l_idx_x[None, None, :] has shape: (1, 1, nx)
        exponent_x = 2j * np.pi * temp_x[..., None] * l_idx_x[None, None, :]
        # Sum over the l-direction for each n and x, resulting shape: (nx, len(x_new))
        varphi_x = np.sum(np.exp(exponent_x), axis=2) / np.sqrt(self.nx * Lx)

        # --- Compose the wavefunction ---
        # Eigenfunction: wf(n,x) = sum_{m} ev[n,m] * varphi_x[m, x]
        eigenfunctions = np.einsum("nm,mx->nx", self.eigenvectors, varphi_x)

        return eigenfunctions
